import numpy as np
import xxhash
from sys import maxsize
from multi_freq_ldpy.long_freq_est.L_GRR import L_GRR_Client
from multi_freq_ldpy.estimators.Histogram_estimator import MI_long, IBU

# [1] Arcolezi et al (2023) "Frequency Estimation of Evolving Data Under Local Differential Privacy" (EDBT 2023).
# [2] Arcolezi et al (2021) "Improving the Utility of Locally Differentially Private Protocols for Longitudinal and Multidimensional Frequency Estimates" (arXiv:2111.04636).

def L_LH_Client(input_data, k, eps_perm, eps_1, optimal=True):
    
    """
    Longitudinal Local Hashing (L-LH) [1] protocol.

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :param optimal: if True, it uses the Optimized Longitudinal LH (L-OLH) protocol from [1];
    :return: tuple of sanitized value and random seed.
    """

    # Validations
    if input_data < 0 or input_data >= k:
        raise ValueError('input_data (integer) should be in the range [0, k-1].')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:
    
        # L-BLH parameter
        g = 2

        # L-OLH parameter
        if optimal:
            alpha = eps_1/eps_perm
            g = int(max(np.rint((np.sqrt(np.exp(4*eps_perm) - 14*np.exp(2*eps_perm) - 12*np.exp(2*eps_perm*(alpha+1)) + 12*np.exp(eps_perm*(alpha+1)) + 12*np.exp(eps_perm*(alpha+3)) + 1) - np.exp(2*eps_perm) + 6*np.exp(eps_perm) - 6*np.exp(eps_perm*alpha) + 1) / (6*(np.exp(eps_perm) - np.exp(eps_perm*alpha)))), 2))

        # Generate random "hash function" (i.e., seed) and hash the user's value
        rnd_seed = np.random.randint(0, maxsize, dtype=np.int64)
        hashed_input_data = (xxhash.xxh32(str(input_data), seed=rnd_seed).intdigest() % g)

        # L-LH perturbation function (i.e., L-GRR-based)
        sanitized_value = L_GRR_Client(hashed_input_data, g, eps_perm, eps_1)

        return (sanitized_value, rnd_seed)

    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')

def L_LH_Aggregator_MI(reports, k, eps_perm, eps_1, optimal=True):
    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all LH-based sanitized values;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :param optimal: if True, it uses the Optimized Longitudinal LH (L-OLH) protocol from [1];
    :return: frequency estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:
    
        # Number of reports
        n = len(reports)
        
        # L-BLH parameter
        g = 2
        
        # L-OLH parameter
        if optimal:
            alpha = eps_1/eps_perm
            g = int(max(np.rint((np.sqrt(np.exp(4*eps_perm) - 14*np.exp(2*eps_perm) - 12*np.exp(2*eps_perm*(alpha+1)) + 12*np.exp(eps_perm*(alpha+1)) + 12*np.exp(eps_perm*(alpha+3)) + 1) - np.exp(2*eps_perm) + 6*np.exp(eps_perm) - 6*np.exp(eps_perm*alpha) + 1) / (6*(np.exp(eps_perm) - np.exp(eps_perm*alpha)))), 2))

        # GRR parameters for round 1
        p1 = np.exp(eps_perm) / (np.exp(eps_perm) + g - 1)
        q1 = (1 - p1) / (g-1)

        # GRR parameters for round 2
        p2 = (q1 - np.exp(eps_1) * p1) / ((-p1 * np.exp(eps_1)) + g*q1*np.exp(eps_1) - q1*np.exp(eps_1) - p1*(g-1)+q1)
        q2 = (1 - p2) / (g-1)
        
        if (np.array([p1, q1, p2, q2]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')
        
        # Count how many times each value has been reported
        obs_count = np.zeros(k)
        for tuple_val_seed in reports:
            for v in range(k):
                if tuple_val_seed[0] == (xxhash.xxh32(str(v), seed=tuple_val_seed[1]).intdigest() % g):
                    obs_count[v] += 1
        
        # Estimate with MI
        q1 = 1 / g  # updating q1 in the server
        norm_est_freq = MI_long(obs_count, n, p1, q1, p2, q2)

        return norm_est_freq
        
    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')
  

def L_LH_Aggregator_IBU(reports, k, eps_perm, eps_1, optimal=True, nb_iter=10000, tol=1e-12, err_func="max_abs"):
    """
    Estimator based on Iterative Bayesian Update[3,4].

    :param reports: list of all LH-based sanitized values;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :param optimal: if True, it uses the Optimized Longitudinal LH (L-OLH) protocol from [1];
    :param nb_iter: number of iterations;
    :param tol: tolerance;
    :param err_func: early stopping function;
    :return: frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:

        # L-BLH parameter
        g = 2

        # L-OLH parameter
        if optimal:
            alpha = eps_1 / eps_perm
            g = int(max(np.rint((np.sqrt(np.exp(4 * eps_perm) - 14 * np.exp(2 * eps_perm) - 12 * np.exp(
                2 * eps_perm * (alpha + 1)) + 12 * np.exp(eps_perm * (alpha + 1)) + 12 * np.exp(
                eps_perm * (alpha + 3)) + 1) - np.exp(2 * eps_perm) + 6 * np.exp(eps_perm) - 6 * np.exp(
                eps_perm * alpha) + 1) / (6 * (np.exp(eps_perm) - np.exp(eps_perm * alpha)))), 2))

        # GRR parameters for round 1
        p1 = np.exp(eps_perm) / (np.exp(eps_perm) + g - 1)
        q1 = (1 - p1) / (g - 1)

        # GRR parameters for round 2
        p2 = (q1 - np.exp(eps_1) * p1) / (
                    (-p1 * np.exp(eps_1)) + g * q1 * np.exp(eps_1) - q1 * np.exp(eps_1) - p1 * (g - 1) + q1)
        q2 = (1 - p2) / (g - 1)

        if (np.array([p1, q1, p2, q2]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # Count how many times each value has been reported
        obs_count = np.zeros(k)
        for tuple_val_seed in reports:
            for v in range(k):
                if tuple_val_seed[0] == (xxhash.xxh32(str(v), seed=tuple_val_seed[1]).intdigest() % g):
                    obs_count[v] += 1
        obs_freq = obs_count / sum(obs_count)

        # Estimate with IBU
        q1 = 1 / g  # updating q1 in the server
        p = p1 * p2 + (1 - p1) * q2  # cf [2]
        q = p1 * q2 + q1 * p2  # cf [2]
        A = np.eye(k)
        A[A == 1] = p
        A[A == 0] = q
        est_freq = IBU(k, A, obs_freq, nb_iter, tol, err_func)

        return est_freq

    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')



import numpy as np
from multi_freq_ldpy.long_freq_est.L_GRR import L_GRR_Client, L_GRR_Aggregator_MI
from multi_freq_ldpy.long_freq_est.L_OUE import L_OUE_Client, L_OUE_Aggregator_MI
from multi_freq_ldpy.long_freq_est.L_OSUE import L_OSUE_Client, L_OSUE_Aggregator_MI
from multi_freq_ldpy.long_freq_est.L_SUE import L_SUE_Client, L_SUE_Aggregator_MI
from multi_freq_ldpy.long_freq_est.L_SOUE import L_SOUE_Client, L_SOUE_Aggregator_MI
from multi_freq_ldpy.long_freq_est.L_ADP import L_ADP_Client, L_ADP_Aggregator_MI
from multi_freq_ldpy.long_freq_est.dBitFlipPM import dBitFlipPM_Client, dBitFlipPM_Aggregator_MI

# [1] Arcolezi et al (2021) "Improving the Utility of Locally Differentially Private Protocols for Longitudinal and Multidimensional Frequency Estimates" (arXiv:2111.04636).
# [2] Erlingsson, Pihur, and Korolova (2014) "RAPPOR: Randomized aggregatable privacy-preserving ordinal response" (ACM CCS).
# [3] Ding, Kulkarni, and Yekhanin (2017) "Collecting telemetry data privately." (NeurIPS).

def SMP_L_GRR_Client(input_tuple, lst_k, d, eps_perm, eps_1):

    """
    Sampling (SMP) a single attribute and using L-GRR [1] protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple (sampled attribute, sanitized value).
    """

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # Report the sampled attribute and its LDP value with L-GRR protocol
    att_sanitized_value = (rnd_att, L_GRR_Client(input_tuple[rnd_att], lst_k[rnd_att], eps_perm, eps_1))

    return att_sanitized_value

def SMP_L_OUE_Client(input_tuple, lst_k, d, eps_perm, eps_1):

    """
    Sampling (SMP) a single attribute and using L-OUE [1] protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple (sampled attribute, sanitized UE vector).
    """

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # Report the sampled attribute and its LDP value with L-OUE protocol
    att_sanitized_value = (rnd_att, L_OUE_Client(input_tuple[rnd_att], lst_k[rnd_att], eps_perm, eps_1))

    return att_sanitized_value

def SMP_L_OSUE_Client(input_tuple, lst_k, d, eps_perm, eps_1):

    """
    Sampling (SMP) a single attribute and using L-OSUE [1] protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple (sampled attribute, sanitized UE vector).
    """

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # Report the sampled attribute and its LDP value with L-OSUE protocol
    att_sanitized_value = (rnd_att, L_OSUE_Client(input_tuple[rnd_att], lst_k[rnd_att], eps_perm, eps_1))

    return att_sanitized_value

def SMP_L_SUE_Client(input_tuple, lst_k, d, eps_perm, eps_1):

    """
    Sampling (SMP) a single attribute and using L-SUE [1] (a.k.a. Basic RAPPOR [2]) protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple (sampled attribute, sanitized UE vector).
    """

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # Report the sampled attribute and its LDP value with L-SUE protocol
    att_sanitized_value = (rnd_att, L_SUE_Client(input_tuple[rnd_att], lst_k[rnd_att], eps_perm, eps_1))

    return att_sanitized_value

def SMP_L_SOUE_Client(input_tuple, lst_k, d, eps_perm, eps_1):

    """
    Sampling (SMP) a single attribute and using L-SOUE [1] protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple (sampled attribute, sanitized UE vector).
    """

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # Report the sampled attribute and its LDP value with L-SOUE protocol
    att_sanitized_value = (rnd_att, L_SOUE_Client(input_tuple[rnd_att], lst_k[rnd_att], eps_perm, eps_1))

    return att_sanitized_value

def SMP_L_ADP_Client(input_tuple, lst_k, d, eps_perm, eps_1):

    """
    Sampling (SMP) a single attribute and using L-ADP [1] protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: tuple (sampled attribute, sanitized value or UE vector).
    """

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # Report the sampled attribute and its LDP value with L-ADP protocol
    att_sanitized_value = (rnd_att, L_ADP_Client(input_tuple[rnd_att], lst_k[rnd_att], eps_perm, eps_1))

    return att_sanitized_value


def SMP_dBitFlipPM_Client(input_tuple, lst_k, lst_b, d_bits, d, eps_perm):
    """
    Sampling (SMP) a single attribute and using dBitFlipPM [3] protocol as local randomizer.

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param lst_b: list of attributes' new domain size (bucketized);
    :param d_bits: number of bits to report;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :return: tuple (sampled attribute, sanitized UE vector).
    """

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # Report the sampled attribute and its LDP value with L-GRR protocol
    att_sanitized_value = (rnd_att, dBitFlipPM_Client(input_tuple[rnd_att], lst_k[rnd_att], lst_b[rnd_att], d_bits, eps_perm))

    return att_sanitized_value


def SMP_L_GRR_Aggregator_MI(reports_tuple, lst_k, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all tuples (sampled attribute, sanitized value);
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_tuple) == 0:

        raise ValueError('List of reports is empty.')

    # Gather users reporting the same attribute
    dic_rep_smp = {att: [] for att in range(d)}
    for val in reports_tuple:
        dic_rep_smp[val[0]].append(val[-1])

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        lst_freq_est.append(L_GRR_Aggregator_MI(dic_rep_smp[idx], lst_k[idx], eps_perm, eps_1))

    return np.array(lst_freq_est, dtype='object')

def SMP_L_OUE_Aggregator_MI(reports_ue_tuple, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_ue_tuple: list of all tuples (sampled attribute, sanitized UE vector);
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_ue_tuple) == 0:
        raise ValueError('List of reports is empty.')

    # Gather users reporting the same attribute
    dic_rep_smp = {att: [] for att in range(d)}
    for val in reports_ue_tuple:
        dic_rep_smp[val[0]].append(val[-1])

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        lst_freq_est.append(L_OUE_Aggregator_MI(dic_rep_smp[idx], eps_perm, eps_1))

    return np.array(lst_freq_est, dtype='object')

def SMP_L_OSUE_Aggregator_MI(reports_ue_tuple, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_ue_tuple: list of all tuples (sampled attribute, sanitized UE vector);
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_ue_tuple) == 0:
        raise ValueError('List of reports is empty.')

    # Gather users reporting the same attribute
    dic_rep_smp = {att: [] for att in range(d)}
    for val in reports_ue_tuple:
        dic_rep_smp[val[0]].append(val[-1])

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        lst_freq_est.append(L_OSUE_Aggregator_MI(dic_rep_smp[idx], eps_perm, eps_1))

    return np.array(lst_freq_est, dtype='object')

def SMP_L_SUE_Aggregator_MI(reports_ue_tuple, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_ue_tuple: list of all tuples (sampled attribute, sanitized UE vector);
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_ue_tuple) == 0:
        raise ValueError('List of reports is empty.')

    # Gather users reporting the same attribute
    dic_rep_smp = {att: [] for att in range(d)}
    for val in reports_ue_tuple:
        dic_rep_smp[val[0]].append(val[-1])

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        lst_freq_est.append(L_SUE_Aggregator_MI(dic_rep_smp[idx], eps_perm, eps_1))

    return np.array(lst_freq_est, dtype='object')

def SMP_L_SOUE_Aggregator_MI(reports_ue_tuple, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_ue_tuple: list of all tuples (sampled attribute, sanitized UE vector);
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_ue_tuple) == 0:
        raise ValueError('List of reports is empty.')

    # Gather users reporting the same attribute
    dic_rep_smp = {att: [] for att in range(d)}
    for val in reports_ue_tuple:
        dic_rep_smp[val[0]].append(val[-1])

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        lst_freq_est.append(L_SOUE_Aggregator_MI(dic_rep_smp[idx], eps_perm, eps_1))

    return np.array(lst_freq_est, dtype='object')
    

def SMP_L_ADP_Aggregator_MI(reports_ue_tuple, lst_k, d, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_ue_tuple: list of all tuples (sampled attribute, sanitized value or UE vector);
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_ue_tuple) == 0:
        raise ValueError('List of reports is empty.')

    # Gather users reporting the same attribute
    dic_rep_smp = {att: [] for att in range(d)}
    for val in reports_ue_tuple:
        dic_rep_smp[val[0]].append(val[-1])

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        lst_freq_est.append(L_ADP_Aggregator_MI(dic_rep_smp[idx], lst_k[idx], eps_perm, eps_1))

    return np.array(lst_freq_est, dtype='object')

def SMP_dBitFlipPM_Aggregator_MI(reports_ue_tuple, lst_b, d_bits, d, eps_perm):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_ue_tuple: list of all tuples (sampled attribute, sanitized UE vector);
    :param d: number of attributes;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_ue_tuple) == 0:
        raise ValueError('List of reports is empty.')

    # Gather users reporting the same attribute
    dic_rep_smp = {att: [] for att in range(d)}
    for val in reports_ue_tuple:
        dic_rep_smp[val[0]].append(val[-1])

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        lst_freq_est.append(dBitFlipPM_Aggregator_MI(dic_rep_smp[idx], lst_b[idx], d_bits, eps_perm))

    return np.array(lst_freq_est, dtype='object')


import numpy as np
from numba import jit

# [1] Ding, Kulkarni, and Yekhanin (2017) "Collecting telemetry data privately" (NeurIPS).

@jit(nopython=True)
def dBitFlipPM_Client(input_data, k, b, d_bits, eps_perm):
    
    """
    dBitFlipPM [1] protocol that applies a single round of sanitization (permanent memoization).

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param b: new domain size to map k to b buckets;
    :param d_bits: number of bits to report;
    :param eps_perm: upper bound of privacy guarantee;
    :return: sanitized UE vector.
    """

    # Validations
    if input_data < 0 or input_data >= k:
        raise ValueError('input_data (integer) should be in the range [0, k-1].')
    if not isinstance(k, int) or not isinstance(b, int) or not isinstance(d_bits, int):
        raise ValueError('k, b, and d_bits need integer values.')
    if b > k or d_bits > b:
        raise ValueError('Please ensure k, b, and d_bits follow d_bits <= b <= k.')
    if eps_perm > 0 and b > 0 and d_bits > 0:

        # SUE parameters
        p1 = np.exp(eps_perm / 2) / (np.exp(eps_perm / 2) + 1)
        q1 = 1 - p1

        # calculate bucket size
        bucket_size = k / b

        # select d_bits random bits j_1, j_2, ..., j_d (without replacement)
        j = np.random.choice(np.arange(b), size=d_bits, replace=False)

        # calculate the buck number of user's value
        bucketized_data = int(input_data / bucket_size)

        # Unary encoding
        permanent_sanitization = np.ones(b) * - 1 # set to -1 non-sampled bits

        # Permanent Memoization
        idx_j = 0
        for i in range(b):
            if i in j: # only the sampled bits
                rand = np.random.random()
                if bucketized_data == j[idx_j]:
                    permanent_sanitization[j[idx_j]] = int(rand <= p1)
                else:
                    permanent_sanitization[j[idx_j]] = int(rand <= q1)

                idx_j+=1

        return permanent_sanitization

    else:
        raise ValueError('Please ensure eps_perm, b, and d_bits have numerical values greater than 0.')

def dBitFlipPM_Aggregator_MI(reports, b, d_bits, eps_perm):
    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all dBitFlipPM sanitized UE vectors;
    :param b: new domain size to map k to b buckets;
    :param d_bits: number of bits to report;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :return: normalized frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(b, int) or not isinstance(d_bits, int):
        raise ValueError('k and d_bits need integer values.')
    if eps_perm > 0:
    
        # Estimated frequency of each bucket
        est_freq = []
        for v in range(b):
            h = 0
            for bi in reports:
                if bi[v] >= 0: # only the sampled bits
                    h += (bi[v] * (np.exp(eps_perm / 2) + 1) - 1) / (np.exp(eps_perm / 2) - 1)
            est_freq.append(h * b / (len(reports) * d_bits ))

        # Ensure non-negativity of estimated frequency
        est_freq = np.array(est_freq).clip(0)

        # Re-normalized estimated frequency
        if sum(est_freq) > 0:
            norm_est_freq = np.nan_to_num(est_freq / sum(est_freq))

        else:
            norm_est_freq = est_freq

        return norm_est_freq

    else:
        raise ValueError('eps_perm needs a numerical value greater than 0.')




# [1] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security 17).

def VAR_Pure(p, q, n=1, f=0):

    """
    Variance value of 'pure' frequency oracles from [1]

    :param p: probability of being 'honest';
    :param q: probability of randomizing value/bit;
    :param n: number of users (if known in advance);
    :param f: real frequency of value (if known in advance - not realistic in locally differentially private scenario);
    :return: variance value.
    """

    return q * (1 - q) / (n * (p - q)**2) + f * (1 - p - q) / (n * (p - q))



import numpy as np
from multi_freq_ldpy.pure_frequency_oracles.GRR import GRR_Client, GRR_Aggregator_MI
from multi_freq_ldpy.pure_frequency_oracles.UE import UE_Client, UE_Aggregator_MI
from multi_freq_ldpy.pure_frequency_oracles.GRR import GRR_Client, GRR_Aggregator_IBU
from multi_freq_ldpy.pure_frequency_oracles.UE import UE_Client, UE_Aggregator_IBU
from multi_freq_ldpy.pure_frequency_oracles.Variance_PURE import VAR_Pure

# [1] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).

def ADP_Client(input_data, k, epsilon, optimal=True):

    """
    Adaptive (ADP) protocol that minimizes variance value (i.e., either GRR or OUE from [1]).

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: sanitized value or UE vector.
    """

    # Validations
    if input_data < 0 or input_data >= k:
        raise ValueError('input_data (integer) should be in the range [0, k-1].')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if epsilon > 0:

        # GRR parameters
        p_grr = np.exp(epsilon) / (np.exp(epsilon) + k - 1)
        q_grr = (1 - p_grr) / (k - 1)

        # Symmetric parameters (p+q = 1)
        p_ue = np.exp(epsilon / 2) / (np.exp(epsilon / 2) + 1)
        q_ue = 1 - p_ue

        # Optimized parameters
        if optimal:
            p_ue = 1 / 2
            q_ue = 1 / (np.exp(epsilon) + 1)

        # Variance values
        var_grr = VAR_Pure(p_grr, q_grr)
        var_ue = VAR_Pure(p_ue, q_ue)

        # Adaptive protocol
        if var_grr <= var_ue:

            return GRR_Client(input_data, k, epsilon)
        else:

            return UE_Client(input_data, k, epsilon, optimal)

    else:
        raise ValueError('k (int) and epsilon (float) need a numerical value.')


def ADP_Aggregator_MI(reports, k, epsilon, optimal=True):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all GRR sanitized values or UE-based vectors;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: normalized frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if epsilon > 0:

        # GRR parameters
        p_grr = np.exp(epsilon) / (np.exp(epsilon) + k - 1)
        q_grr = (1 - p_grr) / (k - 1)

        # Symmetric parameters (p+q = 1)
        p_ue = np.exp(epsilon / 2) / (np.exp(epsilon / 2) + 1)
        q_ue = 1 - p_ue

        # Optimized parameters
        if optimal:
            p_ue = 1 / 2
            q_ue = 1 / (np.exp(epsilon) + 1)

        # Variance values
        var_grr = VAR_Pure(p_grr, q_grr)
        var_ue = VAR_Pure(p_ue, q_ue)

        # Adaptive estimator
        if var_grr <= var_ue:

            return GRR_Aggregator_MI(reports, k, epsilon)
        else:

            return UE_Aggregator_MI(reports, epsilon, optimal)

    else:
        raise ValueError('epsilon needs a numerical value greater than 0.')


def ADP_Aggregator_IBU(reports, k, epsilon, optimal=True, nb_iter=10000, tol=1e-12, err_func="max_abs"):

    """
    Estimator based on Iterative Bayesian Update.

    :param reports: list of all GRR sanitized values or UE-based vectors;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :param nb_iter: number of iterations;
    :param tol: tolerance;
    :param err_func: early stopping function;
    :return: frequency estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or not isinstance(nb_iter, int) or k < 2:
        raise ValueError('k (>=2) and nb_iter need integer values.')
    if nb_iter <= 0 or tol <= 0:
        raise ValueError('nb_iter (int) and tol (float) need values greater than 0')
    if epsilon > 0:

        # GRR parameters
        p_grr = np.exp(epsilon) / (np.exp(epsilon) + k - 1)
        q_grr = (1 - p_grr) / (k - 1)

        # Symmetric parameters (p+q = 1)
        p_ue = np.exp(epsilon / 2) / (np.exp(epsilon / 2) + 1)
        q_ue = 1 - p_ue

        # Optimized parameters
        if optimal:
            p_ue = 1 / 2
            q_ue = 1 / (np.exp(epsilon) + 1)

        # Variance values
        var_grr = VAR_Pure(p_grr, q_grr)
        var_ue = VAR_Pure(p_ue, q_ue)

        # Adaptive estimator
        if var_grr <= var_ue:

            return GRR_Aggregator_IBU(reports, k, epsilon, nb_iter, tol, err_func)
        else:

            return UE_Aggregator_IBU(reports, k, epsilon, optimal, nb_iter, tol, err_func)

    else:
        raise ValueError('epsilon needs a numerical value greater than 0.')



import numpy as np
from numba import jit
from scipy.optimize import minimize_scalar
from multi_freq_ldpy.estimators.Histogram_estimator import MI, IBU

# [1] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).
# [2] Dwork, McSherry, Nissim, and Smith (2006) "Calibrating noise to sensitivity in private data analysis" (TCC).
# [3] Agrawal and Aggarwal (2001,) "On the design and quantification of privacy preserving data mining algorithms" (PODS).
# [4] ElSalamouny and Palamidessi (2020) "Generalized iterative bayesian update and applications to mechanisms for privacy protection" (EuroS&P).

@jit(nopython=True)
def HE_Client(input_data, k, epsilon):
    """
    Histogram encoding (HE) protocol [1].

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :return: sanitized UE-based histogram with Laplace mechanism [2].
    """

    # Validations
    if input_data < 0 or input_data >= k:
        raise ValueError('input_data (integer) should be in the range [0, k-1].')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if epsilon > 0:

        input_ue_data = np.zeros(k)
        input_ue_data[input_data] = 1

        return input_ue_data + np.random.laplace(loc=0.0, scale=2 / epsilon, size=k)

    else:
        raise ValueError('epsilon needs a numerical value greater than 0.')

@jit(nopython=True)
def find_tresh(tresh, epsilon):    
    """
    Objective function for numerical optimization of thresh.

    :param tresh: treshold value for THE [1];
    :param epsilon: privacy guarantee;
    :return: variance (or MSE) when using a given epsilon/tresh with THE [1].
    """
    
    return (2 * (np.exp(epsilon*tresh/2)) - 1) / (1 + (np.exp(epsilon*(tresh-1/2))) - 2*(np.exp(epsilon*tresh/2)))**2

def HE_Aggregator_MI(reports, k, epsilon, use_thresh=True):
    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all GRR-based sanitized values;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param use_thresh: if True will use Thresholding with HE (THE) otherwise Summation with HE (SHE) [1];
    :return: normalized frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if epsilon > 0:
        
        # Number of reports
        n = len(reports)
        
        if use_thresh: # Threshold with HE (THE) [1]
            
            # THE parameters
            thresh = minimize_scalar(find_tresh, bounds=[0.5, 1], method='bounded', args=(epsilon)).x # threshold with optimal value in (0.5, 1) [1]
            p = 1 - 0.5 * np.exp(epsilon*(thresh - 1)/2)
            q = 0.5 * np.exp(-epsilon*thresh/2)

            count_report = np.zeros(k)
            for report in reports:
                ss_the = np.where(report > thresh)[0]
                count_report[ss_the]+=1

            # Estimate with MI
            norm_est_freq = MI(count_report, n, p, q)

            return norm_est_freq 

        else: # Summation with HE (SHE) [1], i.e., Laplace noise has mean=0, so we can just sum up each value
            
            # Ensure non-negativity of estimated frequency
            est_freq = sum(reports).clip(0)            
            
            # Re-normalized estimated frequency
            norm_est_freq = np.nan_to_num(est_freq / sum(est_freq))

            return norm_est_freq   
        
    else:
        raise ValueError('epsilon needs a numerical value greater than 0.')
        
        
def HE_Aggregator_IBU(reports, k, epsilon, use_thresh=True, nb_iter=10000, tol=1e-12, err_func="max_abs"):
    """
    Estimator based on Iterative Bayesian Update[3,4].

    :param reports: list of all GRR-based sanitized values;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param nb_iter: number of iterations;
    :param tol: tolerance;
    :param err_func: early stopping function;
    :return: frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or not isinstance(nb_iter, int) or k < 2:
        raise ValueError('k (>=2) and nb_iter need integer values.')
    if nb_iter <= 0 or tol <= 0:
        raise ValueError('nb_iter (int) and tol (float) need values greater than 0')
    if epsilon > 0:

        if use_thresh:
            # THE parameters
            thresh = minimize_scalar(find_tresh, bounds=[0.5, 1], method='bounded', args=(epsilon)).x # threshold with optimal value in (0.5, 1)
            p = 1 - 0.5 * np.exp(epsilon*(thresh - 1)/2)
            q = 0.5 * np.exp(-epsilon*thresh/2)

            A = np.eye(k)
            A[A == 1] = p
            A[A == 0] = q

            # Count how many times each value has been reported
            obs_count = np.zeros(k)
            for report in reports:
                ss_the = np.where(report > thresh)[0]
                obs_count[ss_the]+=1
            obs_freq = obs_count / sum(obs_count)

            # Estimate with IBU
            est_freq = IBU(k, A, obs_freq, nb_iter, tol, err_func)

            return est_freq

        else:
            raise ValueError('IBU only works with Thresholding HE (THE).')

    else:
        raise ValueError('epsilon needs a numerical value greater than 0.')



# [1] Arcolezi et al (2021) "Improving the Utility of Locally Differentially Private Protocols for Longitudinal and Multidimensional Frequency Estimates" (arXiv:2111.04636).
# [2] Erlingsson, Pihur, and Korolova (2014) "RAPPOR: Randomized aggregatable privacy-preserving ordinal response" (ACM CCS).

def VAR_Long_Pure(p1, q1, p2, q2, n=1, f=0):

    """
    Variance value of longitudinal 'pure' frequency oracles from [1] following the memoization-based framework from [2]

    :param p1: probability of being 'honest' for first round of sanitization (permanent memoization);
    :param q1: probability of randomizing value/bit for round of sanitization (permanent memoization);
    :param p2: probability of being 'honest' for second round of sanitization;
    :param q2: probability of randomizing value/bit for second round of sanitization;
    :param n: number of users (if known in advance);
    :param f: real frequency of value (if known in advance - not realistic in locally differentially private scenario);
    :return: variance value.
    """
    
    sig = f * (2 * p1 * p2 - 2 * p1 * q2 + 2 * q2 - 1) + p2 * q1 + q2 * (1 - q1)

    return sig * (1 - sig) / (n * (p1 - q1)**2 * (p2 - q2)**2)



import numpy as np
from numba import jit
from multi_freq_ldpy.estimators.Histogram_estimator import MI, IBU

# [1] Erlingsson, Pihur, and Korolova (2014) "RAPPOR: Randomized aggregatable privacy-preserving ordinal response" (ACM CCS).
# [2] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).
# [3] Agrawal and Aggarwal (2001,) "On the design and quantification of privacy preserving data mining algorithms" (PODS).
# [4] ElSalamouny and Palamidessi (2020) "Generalized iterative bayesian update and applications to mechanisms for privacy protection" (EuroS&P).

@jit(nopython=True)
def UE_Client(input_data, k, epsilon, optimal=True):
    """
    Unary Encoding (UE) protocol, a.k.a. Basic One-Time RAPPOR (if optimal=False) [1]

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: sanitized UE vector.
    """

    # Validations
    if input_data != None:
        if input_data < 0 or input_data >= k:
            raise ValueError('input_data (integer) should be in the range [0, k-1].')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if epsilon > 0:

        # Symmetric parameters (p+q = 1)
        p = np.exp(epsilon/2) / (np.exp(epsilon/2) + 1)
        q = 1 - p

        # Optimized parameters
        if optimal:
            p = 1 / 2
            q = 1 / (np.exp(epsilon) + 1)

        # Unary encoding
        input_ue_data = np.zeros(k)
        if input_data != None:
            input_ue_data[input_data] = 1

        # Initializing a zero-vector
        sanitized_vec = np.zeros(k)

        # UE perturbation function
        for ind in range(k):
            if input_ue_data[ind] != 1:
                rnd = np.random.random()
                if rnd <= q:
                    sanitized_vec[ind] = 1
            else:
                rnd = np.random.random()
                if rnd <= p:
                    sanitized_vec[ind] = 1
        return sanitized_vec

    else:
        raise ValueError('epsilon (float) needs a numerical value greater than 0.')
        
def UE_Aggregator_MI(reports, epsilon, optimal=True):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all UE-based sanitized vectors;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: normalized frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if epsilon > 0:

        # Number of reports
        n = len(reports)

        # Symmetric parameters (p+q = 1)
        p = np.exp(epsilon/2) / (np.exp(epsilon/2) + 1)
        q = 1 - p

        # Optimized parameters
        if optimal:
            p = 1 / 2
            q = 1 / (np.exp(epsilon) + 1)

        # Estimate with MI
        norm_est_freq = MI(sum(reports), n, p, q)

        return norm_est_freq

    else:
        raise ValueError('epsilon (float) needs a numerical value greater than 0.')


def UE_Aggregator_IBU(reports, k, epsilon, optimal=True, nb_iter=10000, tol=1e-12, err_func="max_abs"):
    """
    Estimator based on Iterative Bayesian Update[3,4].

    :param reports: list of all UE-based sanitized values;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :param nb_iter: number of iterations;
    :param tol: tolerance;
    :param err_func: early stopping function;
    :return: frequency estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or not isinstance(nb_iter, int) or k < 2:
        raise ValueError('k (>=2) and nb_iter need integer values.')
    if nb_iter <= 0 or tol <= 0:
        raise ValueError('nb_iter (int) and tol (float) need values greater than 0')
    if epsilon > 0:

        # Symmetric parameters (p+q = 1)
        p = np.exp(epsilon / 2) / (np.exp(epsilon / 2) + 1)
        q = 1 - p

        # Optimized parameters
        if optimal:
            p = 1 / 2
            q = 1 / (np.exp(epsilon) + 1)

        # UE parameters
        A = np.eye(k)
        A[A == 1] = p
        A[A == 0] = q

        # Count how many times each value was reported
        obs_count = sum(reports)
        obs_freq = obs_count / sum(obs_count)

        # Estimate with IBU
        est_freq = IBU(k, A, obs_freq, nb_iter, tol, err_func)

        return est_freq

    else:
        raise ValueError('epsilon (float) need a numerical value greater than 0.')


import numpy as np
from numba import jit
from multi_freq_ldpy.estimators.Histogram_estimator import MI, IBU

# [1] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).
# [2] Kairouz, Bonawitz, and Ramage (2016) "Discrete distribution estimation under local privacy" (ICML).
# [3] Agrawal and Aggarwal (2001,) "On the design and quantification of privacy preserving data mining algorithms" (PODS).
# [4] ElSalamouny and Palamidessi (2020) "Generalized iterative bayesian update and applications to mechanisms for privacy protection" (EuroS&P).

@jit(nopython=True)
def GRR_Client(input_data, k, epsilon):
    """
    Generalized Randomized Response (GRR) protocol, a.k.a., direct encoding [1] or k-RR [2].

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :return: sanitized value.
    """

    # Validations
    if input_data < 0 or input_data >= k:
        raise ValueError('input_data (integer) should be in the range [0, k-1].')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if epsilon > 0:
        
        # GRR parameters
        p = np.exp(epsilon) / (np.exp(epsilon) + k - 1)

        # Mapping domain size k to the range [0, ..., k-1]
        domain = np.arange(k) 
        
        # GRR perturbation function
        if np.random.binomial(1, p) == 1:
            return input_data

        else:
            return np.random.choice(domain[domain != input_data])

    else:
        raise ValueError('epsilon needs a numerical value greater than 0.')


def GRR_Aggregator_MI(reports, k, epsilon):
    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all GRR-based sanitized values;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :return: normalized frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if epsilon > 0:
            
        # Number of reports
        n = len(reports)

        # GRR parameters
        p = np.exp(epsilon) / (np.exp(epsilon) + k - 1)
        q = (1 - p) / (k - 1)

        # Count how many times each value has been reported
        count_report = np.zeros(k)
        for rep in reports:
            count_report[rep] += 1

        # Estimate with MI
        norm_est_freq = MI(count_report, n, p, q)

        return norm_est_freq

    else:
        raise ValueError('epsilon needs a numerical value greater than 0.')


def GRR_Aggregator_IBU(reports, k, epsilon, nb_iter=10000, tol=1e-12, err_func="max_abs"):
    """
    Estimator based on Iterative Bayesian Update[3,4].

    :param reports: list of all GRR-based sanitized values;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param nb_iter: number of iterations;
    :param tol: tolerance;
    :param err_func: early stopping function;
    :return: frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or not isinstance(nb_iter, int) or k < 2:
        raise ValueError('k (>=2) and nb_iter need integer values.')
    if nb_iter <= 0 or tol <= 0:
        raise ValueError('nb_iter (int) and tol (float) need values greater than 0')
    if epsilon > 0:

        # GRR parameters
        p = np.exp(epsilon) / (np.exp(epsilon) + k - 1)
        q = (1 - p) / (k - 1)
        A = np.eye(k)
        A[A == 1] = p
        A[A == 0] = q

        # Count how many times each value has been reported
        obs_count = np.zeros(k)
        for rep in reports:
            obs_count[rep] += 1
        obs_freq = obs_count / sum(obs_count)

        # Estimate with IBU
        est_freq = IBU(k, A, obs_freq, nb_iter, tol, err_func)

        return est_freq

    else:
        raise ValueError('epsilon needs a numerical value greater than 0.')



import numpy as np
from multi_freq_ldpy.pure_frequency_oracles.GRR import GRR_Client
from multi_freq_ldpy.pure_frequency_oracles.UE import UE_Client
from multi_freq_ldpy.mdim_freq_est.Variance_RSpFD import VAR_RSpFD_GRR, VAR_RSpFD_UE_zero

# [1] Arcolezi et al (2021) "Random Sampling Plus Fake Data: Multidimensional Frequency Estimates With Local Differential Privacy" (ACM CIKM).
# [2] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).

def RSpFD_GRR_Client(input_tuple, lst_k, d, epsilon):

    """
    Random Sampling Plus Fake Data (RS+FD) with Generalized Randomized Response (GRR) protocol as local randomizer [1].

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :return: tuple of sanitized/fake values.
    """

    # Amplified epsilon parameter
    amp_eps = np.log(d * (np.exp(epsilon) - 1) + 1)

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # RS+FD perturbation function
    sanitized_tuple = []
    for idx in range(d):

        # Fake data generation
        if idx != rnd_att:
            sanitized_tuple.append(np.random.randint(lst_k[idx]))

        # Local randomization
        else:
            sanitized_tuple.append(GRR_Client(input_tuple[idx], lst_k[idx], amp_eps))

    return sanitized_tuple

def RSpFD_UE_zero_Client(input_tuple, lst_k, d, epsilon, optimal=True):

    """
    Random Sampling Plus Fake Data (RS+FD) with Unary Encoding (UE) protocol as local randomizer [1].

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: tuple of sanitized/fake UE vectors.
    """

    # Amplified epsilon parameter
    amp_eps = np.log(d * (np.exp(epsilon) - 1) + 1)

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # RS+FD perturbation function
    sanitized_ue_tuple = []
    for idx in range(d):

        # Fake data generation (with UE applied to a zero vector)
        if idx != rnd_att:

            sanitized_ue_tuple.append(UE_Client(None, lst_k[idx], amp_eps, optimal))

        # Local randomization
        else:
            sanitized_ue_tuple.append(UE_Client(input_tuple[idx], lst_k[idx], amp_eps, optimal))

    return sanitized_ue_tuple

def RSpFD_UE_rnd_Client(input_tuple, lst_k, d, epsilon, optimal=True):

    """
    Random Sampling Plus Fake Data (RS+FD) with Unary Encoding (UE) protocol as local randomizer [1].

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: tuple of sanitized/fake UE vectors.
    """

    # Amplified epsilon parameter
    amp_eps = np.log(d * (np.exp(epsilon) - 1) + 1)

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # RS+FD perturbation function
    sanitized_ue_tuple = []
    for idx in range(d):

        # Fake data generation (with UE applied to a random UE vector)
        if idx != rnd_att:
            sanitized_ue_tuple.append(UE_Client(np.random.randint(lst_k[idx]), lst_k[idx], amp_eps, optimal))

        # Local randomization
        else:
            sanitized_ue_tuple.append(UE_Client(input_tuple[idx], lst_k[idx], amp_eps, optimal))

    return sanitized_ue_tuple

def RSpFD_ADP_Client(input_tuple, lst_k, d, epsilon, optimal=True):

    """
    Random Sampling Plus Fake Data (RS+FD) with Adaptive (ADP) protocol (i.e., either RS+FD[GRR] or RS+FD[OUE-z] [1]).

    :param input_tuple: user's true tuple of values;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: tuple of sanitized/fake values or UE vectors.
    """

    # Amplified epsilon parameter
    amp_eps = np.log(d * (np.exp(epsilon) - 1) + 1)

    # Select an attribute at random
    rnd_att = np.random.randint(d)

    # RS+FD perturbation function
    sanitized_tuple = []
    for idx in range(d):
        k = lst_k[idx]

        # GRR parameters with amplified epsilon
        p_grr = np.exp(amp_eps) / (np.exp(amp_eps) + k - 1)
        q_grr = (1 - p_grr) / (k - 1)

        # UE parameters with amplified epsilon
        p_ue = np.exp(amp_eps / 2) / (np.exp(amp_eps / 2) + 1)
        q_ue = 1 - p_ue

        if optimal:
            p_ue = 0.5
            q_ue = 1 / (np.exp(amp_eps) + 1)

        # variance values of using RS+FD[GRR] and RS+FD[OUE-z]
        var_grr = VAR_RSpFD_GRR(p_grr, q_grr, k, d)
        var_ue = VAR_RSpFD_UE_zero(p_ue, q_ue, d)

        # Fake data generation (with either uniform at random or with UE applied to a zero vector)
        if idx != rnd_att:

            # Adaptive fake data generation
            if var_grr <= var_ue:
                sanitized_tuple.append(np.random.randint(k))

            else:
                sanitized_tuple.append(UE_Client(None, k, amp_eps, optimal))

        # Local randomization
        else:

            # Adaptive protocol
            if var_grr <= var_ue:
                sanitized_tuple.append(GRR_Client(input_tuple[idx], k, amp_eps))

            else:
                sanitized_tuple.append(UE_Client(input_tuple[idx], k, amp_eps, optimal))

    return sanitized_tuple

def RSpFD_GRR_Aggregator_MI(reports_tuple, lst_k, d, epsilon):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all RS+FD[GRR]-based sanitized tuples;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_tuple) == 0:

        raise ValueError('List of reports is empty.')

    reports_tuple = np.array(reports_tuple, dtype='object')

    # Number of reports
    n = len(reports_tuple)

    # Amplified epsilon parameter
    amp_eps = np.log(d * (np.exp(epsilon) - 1) + 1)

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        reports = reports_tuple[:, idx]
        k = lst_k[idx]

        # GRR parameters with amplified epsilon
        p = np.exp(amp_eps) / (np.exp(amp_eps) + k - 1)
        q = (1 - p) / (k - 1)

        # Count how many times each value has been reported
        count_report = np.zeros(k)
        for rep in reports:
            count_report[rep] += 1

        # Ensure non-negativity of estimated frequency
        est_freq = np.array(((count_report * d * k) - n * (d - 1 + q * k)) / (n * k * (p - q))).clip(0)

        # Re-normalized estimated frequency
        norm_est_freq = est_freq / sum(est_freq)

        lst_freq_est.append(norm_est_freq)

    return np.array(lst_freq_est, dtype='object')

def RSpFD_UE_zero_Aggregator_MI(reports_tuple, lst_k, d, epsilon, optimal=True):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all RS+FD[UE-zero]-based sanitized tuples;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_tuple) == 0:

        raise ValueError('List of reports is empty.')

    reports_tuple = np.array(reports_tuple, dtype='object')

    # Number of reports
    n = len(reports_tuple)

    # Amplified epsilon parameter
    amp_eps = np.log(d * (np.exp(epsilon) - 1) + 1)

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        reports = reports_tuple[:, idx]
        k = lst_k[idx]

        # UE parameters with amplified epsilon
        p = np.exp(amp_eps / 2) / (np.exp(amp_eps / 2) + 1)
        q = 1 - p

        if optimal:
            p = 1 / 2
            q = 1 / (np.exp(amp_eps) + 1)

        # Ensure non-negativity of estimated frequency
        est_freq = np.array(d*(sum(reports) - n * q) / (n * (p - q))).clip(0)

        # Re-normalized estimated frequency
        norm_est_freq = est_freq / sum(est_freq)

        lst_freq_est.append(norm_est_freq)

    return np.array(lst_freq_est, dtype='object')

def RSpFD_UE_rnd_Aggregator_MI(reports_tuple, lst_k, d, epsilon, optimal=True):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all RS+FD[UE-rnd]-based sanitized tuples;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_tuple) == 0:

        raise ValueError('List of reports is empty.')

    reports_tuple = np.array(reports_tuple, dtype='object')

    # Number of reports
    n = len(reports_tuple)

    # Amplified epsilon parameter
    amp_eps = np.log(d * (np.exp(epsilon) - 1) + 1)

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        reports = reports_tuple[:, idx]
        k = lst_k[idx]

        # UE parameters with amplified epsilon
        p = np.exp(amp_eps / 2) / (np.exp(amp_eps / 2) + 1)
        q = 1 - p

        if optimal:
            p = 1 / 2
            q = 1 / (np.exp(amp_eps) + 1)

        # Ensure non-negativity of estimated frequency
        est_freq = np.array(((sum(reports) * d * k) - n * (q * k + (p - q) * (d - 1) + q * k * (d - 1))) / (n * k * (p - q))).clip(0)

        # Re-normalized estimated frequency
        norm_est_freq = est_freq / sum(est_freq)

        lst_freq_est.append(norm_est_freq)

    return np.array(lst_freq_est, dtype='object')

def RSpFD_ADP_Aggregator_MI(reports_tuple, lst_k, d, epsilon, optimal=True):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) of all d attributes with post-processing to ensure non-negativity.

    :param reports_tuple: list of all RS+FD[GRR]- / RS+FD[UE-zero]-based sanitized tuples;
    :param lst_k: list of attributes' domain size;
    :param d: number of attributes;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized UE (OUE) protocol from [2];
    :return: normalized frequency (histogram) estimation of all d attributes.
    """

    if len(reports_tuple) == 0:

        raise ValueError('List of reports is empty.')

    reports_tuple = np.array(reports_tuple, dtype='object')

    # Number of reports
    n = len(reports_tuple)

    # Amplified epsilon parameter
    amp_eps = np.log(d * (np.exp(epsilon) - 1) + 1)

    # Estimated frequency for all d attributes
    lst_freq_est = []
    for idx in range(d):
        reports = reports_tuple[:, idx]
        k = lst_k[idx]

        # GRR parameters with amplified epsilon
        p_grr = np.exp(amp_eps) / (np.exp(amp_eps) + k - 1)
        q_grr = (1 - p_grr) / (k - 1)

        # UE parameters with amplified epsilon
        p_ue = np.exp(amp_eps / 2) / (np.exp(amp_eps / 2) + 1)
        q_ue = 1 - p_ue

        if optimal:
            p_ue = 0.5
            q_ue = 1 / (np.exp(amp_eps) + 1)

        # Variance values of using RS+FD[GRR] or RS+FD[OUE-z]
        var_grr = VAR_RSpFD_GRR(p_grr, q_grr, k, d)
        var_ue = VAR_RSpFD_UE_zero(p_ue, q_ue, d)

        # Adaptive estimator
        if var_grr <= var_ue:

            # Count how many times each value has been reported
            count_report = np.zeros(k)
            for rep in reports:
                count_report[rep] += 1

            # Ensure non-negativity of estimated frequency
            est_freq = np.array(((count_report * d * k) - n * (d - 1 + q_grr * k)) / (n * k * (p_grr - q_grr))).clip(0)

        else:
            # Ensure non-negativity of estimated frequency
            est_freq = np.array(d * (sum(reports) - n * q_ue) / (n * (p_ue - q_ue))).clip(0)

        # Re-normalized estimated frequency
        norm_est_freq = est_freq / sum(est_freq)

        lst_freq_est.append(norm_est_freq)

    return np.array(lst_freq_est, dtype='object')



import numpy as np
from multi_freq_ldpy.pure_frequency_oracles.GRR import GRR_Client
from multi_freq_ldpy.estimators.Histogram_estimator import MI_long, IBU


# [1] Arcolezi et al (2021) "Improving the Utility of Locally Differentially Private Protocols for Longitudinal and Multidimensional Frequency Estimates" (arXiv:2111.04636).
# [2] Kairouz, Bonawitz, and Ramage (2016) "Discrete distribution estimation under local privacy" (ICML)
# [3] Agrawal and Aggarwal (2001,) "On the design and quantification of privacy preserving data mining algorithms" (PODS).
# [4] ElSalamouny and Palamidessi (2020) "Generalized iterative bayesian update and applications to mechanisms for privacy protection" (EuroS&P).

# The analytical analysis of how to calculate parameters (p1, q2, p2, q2) is from: https://github.com/hharcolezi/ldp-protocols-mobility-cdrs/blob/main/papers/%5B4%5D/1_ALLOMFREE_Analysis.ipynb

def L_GRR_Client(input_data, k, eps_perm, eps_1):

    """
    Longitudinal GRR (L-GRR) [1] protocol that chaines GRR [2] for both first and second rounds of sanitization.

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: sanitized value.
    """

    # Validations
    if input_data < 0 or input_data >= k:
        raise ValueError('input_data (integer) should be in the range [0, k-1].')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:

        # GRR parameters for round 1
        p1 = np.exp(eps_perm) / (np.exp(eps_perm) + k - 1)
        q1 = (1 - p1) / (k - 1)

        # GRR parameters for round 2
        p2 = (q1 - np.exp(eps_1) * p1) / ((-p1 * np.exp(eps_1)) + k*q1*np.exp(eps_1) - q1*np.exp(eps_1) - p1*(k-1)+q1)
        q2 = (1 - p2) / (k-1)
        
        if (np.array([p1, q1, p2, q2]) >= 0).all():
            pass
        else: 
            raise ValueError('Probabilities are negative, selecting eps_1 << eps_perm might probably solve it.')

        # Get epsilon of second round of sanitization
        eps_sec_round = np.log(p2 / q2)

        # First round of sanitization (permanent memoization) with GRR using user's input_data
        first_sanitization = GRR_Client(input_data, k, eps_perm)

        # Second round of sanitization with GRR using first_sanitization as input
        second_sanitization = GRR_Client(first_sanitization, k, eps_sec_round)
        
        return second_sanitization

    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')

def L_GRR_Aggregator_MI(reports, k, eps_perm, eps_1):
    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all L-GRR sanitized values;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:

        # Number of reports
        n = len(reports)

        # GRR parameters for round 1
        p1 = np.exp(eps_perm) / (np.exp(eps_perm) + k - 1)
        q1 = (1 - p1) / (k - 1)

        # GRR parameters for round 2
        p2 = (q1 - np.exp(eps_1) * p1) / ((-p1 * np.exp(eps_1)) + k*q1*np.exp(eps_1) - q1*np.exp(eps_1) - p1*(k-1)+q1)
        q2 = (1 - p2) / (k-1)

        if (np.array([p1, q1, p2, q2]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # Count how many times each value has been reported
        count_report = np.zeros(k)
        for rep in reports:
            count_report[rep] += 1

        # Estimate with MI
        norm_est_freq = MI_long(count_report, n, p1, q1, p2, q2)

        return norm_est_freq

    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')

def L_GRR_Aggregator_IBU(reports, k, eps_perm, eps_1, nb_iter=10000, tol=1e-12, err_func="max_abs"):
    """
    Estimator based on Iterative Bayesian Update[3,4].

    :param reports: list of all L-GRR sanitized values;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :param nb_iter: number of iterations;
    :param tol: tolerance;
    :param err_func: early stopping function;
    :return: frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or not isinstance(nb_iter, int) or k < 2:
        raise ValueError('k (>=2) and nb_iter need integer values.')
    if nb_iter <= 0 or tol <= 0:
        raise ValueError('nb_iter (int) and tol (float) need values greater than 0')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:

        # GRR parameters for round 1
        p1 = np.exp(eps_perm) / (np.exp(eps_perm) + k - 1)
        q1 = (1 - p1) / (k - 1)

        # GRR parameters for round 2
        p2 = (q1 - np.exp(eps_1) * p1) / (
                    (-p1 * np.exp(eps_1)) + k * q1 * np.exp(eps_1) - q1 * np.exp(eps_1) - p1 * (k - 1) + q1)
        q2 = (1 - p2) / (k - 1)

        if (np.array([p1, q1, p2, q2]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # Count how many times each value has been reported
        obs_count = np.zeros(k)
        for rep in reports:
            obs_count[rep] += 1
        obs_freq = obs_count / len(reports)

        # Estimate with IBU
        p = p1 * p2 + (1 - p1) * q2  # cf [1]
        q = p1 * q2 + q1 * p2  # cf [1]
        A = np.eye(k)
        A[A == 1] = p
        A[A == 0] = q
        est_freq = IBU(k, A, obs_freq, nb_iter, tol, err_func)

        return est_freq

    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')



import numpy as np
from numba import jit
from multi_freq_ldpy.estimators.Histogram_estimator import MI_long, IBU

# [1] Arcolezi et al (2021) "Improving the Utility of Locally Differentially Private Protocols for Longitudinal and Multidimensional Frequency Estimates" (arXiv:2111.04636).
# [2] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).
# [3] Agrawal and Aggarwal (2001,) "On the design and quantification of privacy preserving data mining algorithms" (PODS).
# [4] ElSalamouny and Palamidessi (2020) "Generalized iterative bayesian update and applications to mechanisms for privacy protection" (EuroS&P).

# The analytical analysis of how to calculate parameters (p1, q2, p2, q2) is from: https://github.com/hharcolezi/ldp-protocols-mobility-cdrs/blob/main/papers/%5B4%5D/1_ALLOMFREE_Analysis.ipynb

@jit(nopython=True)
def L_OSUE_Client(input_data, k, eps_perm, eps_1):

    """
    Longitudinal OUE-SUE (L-OSUE) [1] protocol that chains OUE [2] for first round and SUE [2] for second round of sanitization.

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: sanitized UE vector.
    """

    # Validations
    if input_data < 0 or input_data >= k:
        raise ValueError('input_data (integer) should be in the range [0, k-1].')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:

        # OUE parameters for round 1
        p1 = 1 / 2
        q1 = 1 / (np.exp(eps_perm) + 1)

        # SUE parameters for round 2
        p2 = (1 - np.exp(eps_1 + eps_perm)) / (np.exp(eps_1) - np.exp(eps_perm) - np.exp(eps_1 + eps_perm) + 1)
        q2 = 1 - p2

        if (np.array([p1, q1, p2, q2]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # Unary encoding
        input_ue_data = np.zeros(k)
        if input_data != None:
            input_ue_data[input_data] = 1

        # First round of sanitization (permanent memoization) with OUE using user's input_ue_data
        first_sanitization = np.zeros(k)
        for ind in range(k):
            if input_ue_data[ind] != 1:
                rnd = np.random.random()
                if rnd <= q1:
                    first_sanitization[ind] = 1
            else:
                rnd = np.random.random()
                if rnd <= p1:
                    first_sanitization[ind] = 1

        # Second round of sanitization with SUE using first_sanitization as input
        second_sanitization = np.zeros(k)
        for ind in range(k):
            if first_sanitization[ind] != 1:
                rnd = np.random.random()
                if rnd <= q2:
                    second_sanitization[ind] = 1
            else:
                rnd = np.random.random()
                if rnd <= p2:
                    second_sanitization[ind] = 1

        return second_sanitization

    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')


def L_OSUE_Aggregator_MI(reports, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all L-OSUE sanitized UE-based vectors;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:

        # Number of reports
        n = len(reports)

        # OUE parameters for round 1
        p1 = 1 / 2
        q1 = 1 / (np.exp(eps_perm) + 1)

        # SUE parameters for round 2
        p2 = (1 - np.exp(eps_1 + eps_perm)) / (np.exp(eps_1) - np.exp(eps_perm) - np.exp(eps_1 + eps_perm) + 1)
        q2 = 1 - p2

        if (np.array([p1, q1, p2, q2]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # Count how many times each bit has been reported
        count_report = sum(reports)

        # Estimate with MI
        norm_est_freq = MI_long(count_report, n, p1, q1, p2, q2)

        return norm_est_freq

    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')


def L_OSUE_Aggregator_IBU(reports, k, eps_perm, eps_1, nb_iter=10000, tol=1e-12, err_func="max_abs"):
    """
    Estimator based on Iterative Bayesian Update[3,4].

    :param reports: list of all L-OSUE UE-based vectors;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :param nb_iter: number of iterations;
    :param tol: tolerance;
    :param err_func: early stopping function;
    :return: frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or not isinstance(nb_iter, int) or k < 2:
        raise ValueError('k (>=2) and nb_iter need integer values.')
    if nb_iter <= 0 or tol <= 0:
        raise ValueError('nb_iter (int) and tol (float) need values greater than 0')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:

        # OUE parameters for round 1
        p1 = 1 / 2
        q1 = 1 / (np.exp(eps_perm) + 1)

        # SUE parameters for round 2
        p2 = (1 - np.exp(eps_1 + eps_perm)) / (np.exp(eps_1) - np.exp(eps_perm) - np.exp(eps_1 + eps_perm) + 1)
        q2 = 1 - p2

        if (np.array([p1, q1, p2, q2]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # Count how many times each bit has been reported
        count_report = sum(reports)
        obs_freq = count_report/sum(count_report)

        # Estimate with IBU
        p = p1 * p2 + (1 - p1) * q2  # cf [1]
        q = (1 - q1) * q2 + q1 * p2  # cf [1]
        A = np.eye(k)
        A[A == 1] = p
        A[A == 0] = q
        est_freq = IBU(k, A, obs_freq, nb_iter, tol, err_func)

        return est_freq

    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')





import numpy as np
from numba import jit
from multi_freq_ldpy.estimators.Histogram_estimator import MI, IBU

# [1] Ye and Barg (2018) "Optimal schemes for discrete distribution estimation under locally differential privacy" (IEEE Transactions on Information Theory).
# [2] Wang et al (2016) "Mutual information optimally local private discrete distribution estimation" (arXiv:1607.08025).
# [3] Agrawal and Aggarwal (2001,) "On the design and quantification of privacy preserving data mining algorithms" (PODS).
# [4] ElSalamouny and Palamidessi (2020) "Generalized iterative bayesian update and applications to mechanisms for privacy protection" (EuroS&P).

@jit(nopython=True)
def SS_Client(input_data, k, epsilon):
    """
    Subset Selection (SS) protocol [1,2].

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :return: set of sub_k sanitized values.
    """

    # Validations
    if input_data < 0 or input_data >= k:
        raise ValueError('input_data (integer) should be in the range [0, k-1].')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if epsilon > 0:

        # Mapping domain size k to the range [0, ..., k-1]
        domain = np.arange(k)

        # SS parameters
        sub_k = int(max(1, np.rint(k / (np.exp(epsilon) + 1))))
        p_v = sub_k * np.exp(epsilon) / (sub_k * np.exp(epsilon) + k - sub_k)

        # SS perturbation function
        rnd = np.random.random()
        sub_set = np.zeros(sub_k, dtype='int64')
        if rnd <= p_v:
            sub_set[0] = int(input_data)
            sub_set[1:] = np.random.choice(domain[domain != input_data], size=sub_k-1, replace=False)
            return sub_set

        else:
            return np.random.choice(domain[domain != input_data], size=sub_k, replace=False)

    else:
        raise ValueError('epsilon (float) needs a numerical value greater than 0.')

def SS_Aggregator_MI(reports, k, epsilon):
    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all GRR-based sanitized values;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :return: normalized frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if epsilon > 0:
            
        # Number of reports
        n = len(reports)

        # SS parameters
        sub_k = int(max(1, np.rint(k / (np.exp(epsilon) + 1))))
        p = sub_k * np.exp(epsilon) / (sub_k * np.exp(epsilon) + k - sub_k)
        q = ((sub_k - 1) * (sub_k * np.exp(epsilon)) + (k - sub_k) * sub_k) / ((k - 1) * (sub_k * np.exp(epsilon) + k - sub_k))

        # Count how many times each value has been reported
        count_report = np.zeros(k)
        for rep in reports:
            for i in range(sub_k):
                count_report[rep[i]] += 1

        # Estimate with MI
        norm_est_freq = MI(count_report, n, p, q)

        return norm_est_freq

    else:
        raise ValueError('epsilon (float) needs a numerical value greater than 0.')


def SS_Aggregator_IBU(reports, k, epsilon, nb_iter=10000, tol=1e-12, err_func="max_abs"):
    """
    Estimator based on Iterative Bayesian Update[3,4].

    :param reports: list of all SS-based sanitized values;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param nb_iter: number of iterations;
    :param tol: tolerance;
    :param err_func: early stopping function;
    :return: frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or not isinstance(nb_iter, int) or k < 2:
        raise ValueError('k (>=2) and nb_iter need integer values.')
    if nb_iter <= 0 or tol <= 0:
        raise ValueError('nb_iter (int) and tol (float) need values greater than 0')
    if epsilon > 0:

        # Number of reports
        n = len(reports)

        # SS parameters
        sub_k = int(max(1, np.rint(k / (np.exp(epsilon) + 1))))
        p = sub_k * np.exp(epsilon) / (sub_k * np.exp(epsilon) + k - sub_k)
        q = ((sub_k - 1) * (sub_k * np.exp(epsilon)) + (k - sub_k) * sub_k) / (
                    (k - 1) * (sub_k * np.exp(epsilon) + k - sub_k))
        A = np.eye(k)
        A[A == 1] = p
        A[A == 0] = q

        # Count how many times each value has been reported
        obs_count = np.zeros(k)
        for rep in reports:
            for i in range(sub_k):
                obs_count[rep[i]] += 1
        obs_freq = obs_count / sum(obs_count)

        # Estimate with IBU
        est_freq = IBU(k, A, obs_freq, nb_iter, tol, err_func)

        return est_freq

    else:
        raise ValueError('epsilon (float) need a numerical value greater than 0.')


import numpy as np
from numba import jit

# [1] Agrawal et al (2001) "On the Design and Quantification of Privacy Preserving Data Mining Algorithms" (PODS).
# [2] Ehab et al (2020) "Generalized Iterative Bayesian Update and Applications to Mechanisms for Privacy Protection" (EuroS&P).

def MI(count_report, n, p, q):
    """
    Matrix Inversion (MI).

    :param count_report : number of times that each value was reported;
    :param n : number of reports;
    :param p : probability of being honest;
    :param q : probability of lying;
    :return : normalized frequency (histogram) estimation.
    """

    # # Validations
    # if len(count_report) == 0:
    #     raise ValueError('List of count_report is empty.')
    # if not isinstance(n, int) or not isinstance(p, float) or not isinstance(q, float):
    #     raise ValueError('n (int), p (float), q (float) need numerical values.')

    # Ensure non-negativity of estimated frequency
    est_freq = np.array((count_report - n * q) / (p - q)).clip(0)

    # Re-normalized estimated frequency
    if sum(est_freq) > 0:
        norm_est_freq = np.nan_to_num(est_freq / sum(est_freq))

    else:
        norm_est_freq = est_freq

    return norm_est_freq

def MI_long(count_report, n, p1, q1, p2, q2):
    """
    Matrix Inversion (MI).

    :param count_report : number of times that each value was reported;
    :param n : number of reports;
    :param p1 : probability of being honest in sanitization round 1;
    :param q1 : probability of lying in sanitization round 1;
    :param p2 : probability of being honest in sanitization round 2;
    :param q2 : probability of lying in sanitization round 2;
    :return : normalized frequency (histogram) estimation.
    """

    # # Validations
    # if len(count_report) == 0:
    #     raise ValueError('List of count_report is empty.')
    # if not isinstance(n, int) or not isinstance(p, float) or not isinstance(q, float):
    #     raise ValueError('n (int), p (float), q (float) need numerical values.')

    # Ensure non-negativity of estimated frequency
    est_freq = ((count_report - n * q1 * (p2 - q2) - n * q2) / (n * (p1 - q1) * (p2 - q2))).clip(0)

    # Re-normalized estimated frequency
    if sum(est_freq) > 0:
        norm_est_freq = np.nan_to_num(est_freq / sum(est_freq))

    else:
        norm_est_freq = est_freq

    return norm_est_freq


@jit(nopython=True)
def IBU(k, A, obs_freq, nb_iter, tol, err_func):
    """
    Iterative Bayesian Update (IBU)[1,2].

    :param k : attribute's domain size;
    :param A : probability matrix;
    :param obs_freq : observed frequency;
    :param nb_iter : number of iterations;
    :param tol : tolerance;
    :param err_func: early stopping function;
    :return : frequency estimation.
    """

    # # Validations
    # if len(obs_freq) == 0:
    #     raise ValueError('List of obs_freq is empty.')
    # if not all(isinstance(l, np.ndarray) for l in A):
    #     raise ValueError('A must be a np.ndarray.')
    # if not isinstance(k, int) or not isinstance(nb_iter, int):
    #     raise ValueError('k and nb_iter need integer values.')
    # if nb_iter<=0 or tol<=0:
    #     raise ValueError('nb_iter (int) and tol (float) need values greater than 0')

    # Step 1 - Expectation: initialization of probabilities as a uniform distribution
    est_freq = np.ones(k) / k
    est_freq_t = None

    # Step 2 - Maximization: calculating estimated frequencies
    if err_func=="max_abs":
        for _ in range(nb_iter):
            est_freq_t = est_freq * np.dot(A, obs_freq / np.dot(A, est_freq))
            if np.abs(est_freq - est_freq_t).max() < tol:
                return est_freq_t
            else:
                est_freq = est_freq_t

    elif err_func=="mse":
        for _ in range(nb_iter):
            est_freq_t = est_freq * np.dot(A, obs_freq / np.dot(A, est_freq))
            if np.square(np.subtract(est_freq_t, est_freq)).mean() < tol:
                return est_freq_t
            else:
                est_freq = est_freq_t

    elif err_func == "mae":
        for _ in range(nb_iter):
            est_freq_t = est_freq * np.dot(A, obs_freq / np.dot(A, est_freq))
            if np.abs(est_freq - est_freq_t).mean() < tol:
                return est_freq_t
            else:
                est_freq = est_freq_t

    elif err_func == "max_squared":
        for _ in range(nb_iter):
            est_freq_t = est_freq * np.dot(A, obs_freq / np.dot(A, est_freq))
            if np.square(np.subtract(est_freq_t, est_freq)).max() < tol:
                return est_freq_t
            else:
                est_freq = est_freq_t
    else:
        raise ValueError('Error function unknown. Options are: max_abs, mse, mae, max_squared.')

    return est_freq_t




import numpy as np
from multi_freq_ldpy.long_freq_est.L_GRR import L_GRR_Client, L_GRR_Aggregator_MI
from multi_freq_ldpy.long_freq_est.L_OSUE import L_OSUE_Client, L_OSUE_Aggregator_MI
from multi_freq_ldpy.long_freq_est.Variance_LONG_PURE import VAR_Long_Pure

# [1] Arcolezi et al (2021) "Improving the Utility of Locally Differentially Private Protocols for Longitudinal and Multidimensional Frequency Estimates" (arXiv:2111.04636).

# The analytical analysis of how to calculate parameters (p1, q2, p2, q2) is from: https://github.com/hharcolezi/ldp-protocols-mobility-cdrs/blob/main/papers/%5B4%5D/1_ALLOMFREE_Analysis.ipynb

def L_ADP_Client(input_data, k, eps_perm, eps_1):

    """
    Longitudinal Adaptive (L-ADP) protocol that minimizes variance value (i.e., either L-GRR or L-OSUE from [1], a.k.a. ALLOMFREE).

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: sanitized value or UE vector.
    """

    # Validations
    if input_data < 0 or input_data >= k:
        raise ValueError('input_data (integer) should be in the range [0, k-1].')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:

        # GRR parameters for round 1
        p1_grr = np.exp(eps_perm) / (np.exp(eps_perm) + k - 1)
        q1_grr = (1 - p1_grr) / (k - 1)

        #  GRR parameters for round 2
        p2_grr = (q1_grr - np.exp(eps_1) * p1_grr) / (
                    (-p1_grr * np.exp(eps_1)) + k * q1_grr * np.exp(eps_1) - q1_grr * np.exp(eps_1) - p1_grr * (k - 1) + q1_grr)
        q2_grr = (1 - p2_grr) / (k - 1)

        if (np.array([p1_grr, q1_grr, p2_grr, q2_grr]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # OUE parameters for round 1
        p1_ue = 1 / 2
        q1_ue = 1 / (np.exp(eps_perm) + 1)

        # SUE parameters for round 2
        p2_ue = (1 - np.exp(eps_1 + eps_perm)) / (np.exp(eps_1) - np.exp(eps_perm) - np.exp(eps_1 + eps_perm) + 1)
        q2_ue = 1 - p2_ue

        if (np.array([p1_ue, q1_ue, p2_ue, q2_ue]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # Variance values of L-GRR and L-OSUE
        var_l_grr = VAR_Long_Pure(p1_grr, q1_grr, p2_grr, q2_grr)
        var_l_osue = VAR_Long_Pure(p1_ue, q1_ue, p2_ue, q2_ue)

        # Adaptive longitudinal protocol (a.k.a. ALLOMFREE in [1])
        if var_l_grr <= var_l_osue:

            return L_GRR_Client(input_data, k, eps_perm, eps_1)
        else:

            return L_OSUE_Client(input_data, k, eps_perm, eps_1)

    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')


def L_ADP_Aggregator_MI(reports, k, eps_perm, eps_1):

    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all sanitized L-GRR values or L-OSUE UE-based vectors;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :return: normalized frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:

        # GRR parameters for round 1
        p1_grr = np.exp(eps_perm) / (np.exp(eps_perm) + k - 1)
        q1_grr = (1 - p1_grr) / (k - 1)

        #  GRR parameters for round 2
        p2_grr = (q1_grr - np.exp(eps_1) * p1_grr) / (
                (-p1_grr * np.exp(eps_1)) + k * q1_grr * np.exp(eps_1) - q1_grr * np.exp(eps_1) - p1_grr * (k - 1) + q1_grr)
        q2_grr = (1 - p2_grr) / (k - 1)

        if (np.array([p1_grr, q1_grr, p2_grr, q2_grr]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # OUE parameters for round 1
        p1_ue = 1 / 2
        q1_ue = 1 / (np.exp(eps_perm) + 1)

        # SUE parameters for round 2
        p2_ue = (1 - np.exp(eps_1 + eps_perm)) / (np.exp(eps_1) - np.exp(eps_perm) - np.exp(eps_1 + eps_perm) + 1)
        q2_ue = 1 - p2_ue

        if (np.array([p1_ue, q1_ue, p2_ue, q2_ue]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # Variance values of L-GRR and L-OSUE
        var_l_grr = VAR_Long_Pure(p1_grr, q1_grr, p2_grr, q2_grr)
        var_l_osue = VAR_Long_Pure(p1_ue, q1_ue, p2_ue, q2_ue)

        # Adaptive estimator
        if var_l_grr <= var_l_osue:

            return L_GRR_Aggregator_MI(reports, k, eps_perm, eps_1)

        else:
            return L_OSUE_Aggregator_MI(reports, eps_perm, eps_1)

    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')

def L_ADP_Aggregator_IBU(reports, k, eps_perm, eps_1, nb_iter=10000, tol=1e-12, err_func="max_abs"):
    """
    Estimator based on Iterative Bayesian Update[3,4].

    :param reports: list of all L-GRR sanitized values or UE-based vectors;
    :param k: attribute's domain size;
    :param eps_perm: upper bound of privacy guarantee (infinity reports);
    :param eps_1: lower bound of privacy guarantee (a single report), thus, eps_1 < eps_perm;
    :param nb_iter: number of iterations;
    :param tol: tolerance;
    :param err_func: early stopping function;
    :return: frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or not isinstance(nb_iter, int) or k < 2:
        raise ValueError('k (>=2) and nb_iter need integer values.')
    if nb_iter <= 0 or tol <= 0:
        raise ValueError('nb_iter (int) and tol (float) need values greater than 0')
    if eps_perm < 0 or eps_1 < 0:
        raise ValueError('Please ensure eps_perm and eps_1 have numerical values greater than 0.')
    if eps_1 < eps_perm:

        # GRR parameters for round 1
        p1_grr = np.exp(eps_perm) / (np.exp(eps_perm) + k - 1)
        q1_grr = (1 - p1_grr) / (k - 1)

        #  GRR parameters for round 2
        p2_grr = (q1_grr - np.exp(eps_1) * p1_grr) / (
                (-p1_grr * np.exp(eps_1)) + k * q1_grr * np.exp(eps_1) - q1_grr * np.exp(eps_1) - p1_grr * (
                    k - 1) + q1_grr)
        q2_grr = (1 - p2_grr) / (k - 1)

        if (np.array([p1_grr, q1_grr, p2_grr, q2_grr]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # OUE parameters for round 1
        p1_ue = 1 / 2
        q1_ue = 1 / (np.exp(eps_perm) + 1)

        # SUE parameters for round 2
        p2_ue = (1 - np.exp(eps_1 + eps_perm)) / (np.exp(eps_1) - np.exp(eps_perm) - np.exp(eps_1 + eps_perm) + 1)
        q2_ue = 1 - p2_ue

        if (np.array([p1_ue, q1_ue, p2_ue, q2_ue]) >= 0).all():
            pass
        else:
            raise ValueError('Probabilities are negative.')

        # Variance values of L-GRR and L-OSUE
        var_l_grr = VAR_Long_Pure(p1_grr, q1_grr, p2_grr, q2_grr)
        var_l_osue = VAR_Long_Pure(p1_ue, q1_ue, p2_ue, q2_ue)

        # Adaptive estimator
        if var_l_grr <= var_l_osue:

            return L_GRR_Aggregator_IBU(reports, k, eps_perm, eps_1, nb_iter, tol, err_func)

        else:
            return L_OSUE_Aggregator_IBU(reports, eps_perm, eps_1, nb_iter, tol, err_func)
    else:
        raise ValueError('Please set eps_1 (single report, i.e., lower bound) < eps_perm (infinity reports, i.e., upper bound)')



import numpy as np
from sys import maxsize
import xxhash
from multi_freq_ldpy.estimators.Histogram_estimator import MI, IBU
from multi_freq_ldpy.pure_frequency_oracles.GRR import GRR_Client

# [1] Wang et al (2017) "Locally differentially private protocols for frequency estimation" (USENIX Security).
# [2] Bassily and Smith "Local, private, efficient protocols for succinct histograms" (STOC).
# [3] Agrawal and Aggarwal (2001,) "On the design and quantification of privacy preserving data mining algorithms" (PODS).
# [4] ElSalamouny and Palamidessi (2020) "Generalized iterative bayesian update and applications to mechanisms for privacy protection" (EuroS&P).

# Code adapted from pure-ldp library (https://github.com/Samuel-Maddock/pure-LDP) developed by Samuel Maddock

def LH_Client(input_data, k, epsilon, optimal=True):
    
    """
    Local Hashing (LH) protocol[1], which is logically equivalent to the random matrix projection technique in [2].

    :param input_data: user's true value;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized LH (OLH) protocol from [1];
    :return: tuple of sanitized value and random seed.
    """

    # Validations
    if input_data < 0 or input_data >= k:
        raise ValueError('input_data (integer) should be in the range [0, k-1].')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if epsilon > 0:

        # Binary LH (BLH) parameter
        g = 2

        # Optimal LH (OLH) parameter
        if optimal:
            g = int(round(np.exp(epsilon))) + 1

        # Generate random seed and hash the user's value
        rnd_seed = np.random.randint(0, maxsize, dtype=np.int64)
        hashed_input_data = (xxhash.xxh32(str(input_data), seed=rnd_seed).intdigest() % g)

        # LH perturbation function (i.e., GRR-based)
        sanitized_value = GRR_Client(hashed_input_data, g, epsilon)

        return (sanitized_value, rnd_seed)

    else:
        raise ValueError('epsilon (float) needs a numerical value greater than 0.')

def LH_Aggregator_MI(reports, k, epsilon, optimal=True):
    """
    Statistical Estimator for Normalized Frequency (0 -- 1) with post-processing to ensure non-negativity.

    :param reports: list of all LH-based sanitized values;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized LH (OLH) protocol from [1];
    :return: frequency estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or k < 2:
        raise ValueError('k needs an integer value >=2.')
    if epsilon > 0:
            
        # Number of reports
        n = len(reports)

        # Binary LH (BLH) parameter
        g = 2

        # Optimal LH (OLH) parameter
        if optimal:
            g = int(round(np.exp(epsilon))) + 1

        # Count how many times each value has been reported
        count_report = np.zeros(k)
        for tuple_val_seed in reports:
            for v in range(k):
                if tuple_val_seed[0] == (xxhash.xxh32(str(v), seed=tuple_val_seed[1]).intdigest() % g):
                    count_report[v] += 1

        # Estimate with MI
        p = np.exp(epsilon) / (np.exp(epsilon) + g - 1) # GRR parameters with reduced domain size g
        q = 1 / g # g value differs at the server side [1]
        norm_est_freq = MI(count_report, n, p, q)

        return norm_est_freq

    else:
        raise ValueError('epsilon (float) needs a numerical value greater than 0.')


def LH_Aggregator_IBU(reports, k, epsilon, optimal=True, nb_iter=10000, tol=1e-12, err_func="max_abs"):
    """
    Estimator based on Iterative Bayesian Update[3,4].

    :param reports: list of all LH-based sanitized values;
    :param k: attribute's domain size;
    :param epsilon: privacy guarantee;
    :param optimal: if True, it uses the Optimized LH (OLH) protocol from [1];
    :param nb_iter: number of iterations;
    :param tol: tolerance;
    :param err_func: early stopping function;
    :return: frequency (histogram) estimation.
    """

    # Validations
    if len(reports) == 0:
        raise ValueError('List of reports is empty.')
    if not isinstance(k, int) or not isinstance(nb_iter, int) or k < 2:
        raise ValueError('k (>=2) and nb_iter need integer values.')
    if nb_iter <= 0 or tol <= 0:
        raise ValueError('nb_iter (int) and tol (float) need values greater than 0')
    if epsilon > 0:

        # Binary LH (BLH) parameter
        g = 2

        # Optimal LH (OLH) parameter
        if optimal:
            g = int(round(np.exp(epsilon))) + 1

        # Count how many times each value has been reported
        obs_count = np.zeros(k)
        for tuple_val_seed in reports:
            for v in range(k):
                if tuple_val_seed[0] == (xxhash.xxh32(str(v), seed=tuple_val_seed[1]).intdigest() % g):
                    obs_count[v] += 1
        obs_freq = obs_count / sum(obs_count)

        # Estimate with IBU
        p = np.exp(epsilon) / (np.exp(epsilon) + g - 1) # GRR parameters with reduced domain size g
        q = 1 / g  # g value differs at the server side [1]
        A = np.eye(k)
        A[A == 1] = p
        A[A == 0] = q
        est_freq = IBU(k, A, obs_freq, nb_iter, tol, err_func)

        return est_freq

    else:
        raise ValueError('epsilon (float) need a numerical value greater than 0.')



# [1] Arcolezi et al (2021) "Random Sampling Plus Fake Data: Multidimensional Frequency Estimates With Local Differential Privacy" (ACM CIKM).

def VAR_RSpFD_GRR(p, q, k, d, n=1, f=0):

    """
    Variance value of using RS+FD[GRR] protocol of [1]

    :param p: probability of being 'honest';
    :param q: probability of randomizing value/bit;
    :param k: attribute's domain size;
    :param d: number of attributes;
    :param n: number of users (if known in advance);
    :param f: real frequency of value (if known in advance - not realistic in locally differentially private scenario);
    :return: variance value.
    """

    sig_grr = (1 / d) * (q + f * (p - q) + (d - 1) / k)

    var_grr = ((d**2 * sig_grr * (1 - sig_grr)) / (n * (p - q)**2))

    return var_grr

def VAR_RSpFD_UE_zero(p, q, d, n=1, f=0):

    """
    Variance value of using RS+FD[UE-zero] protocol of [1]

    :param p: probability of being 'honest';
    :param q: probability of randomizing value/bit;
    :param d: number of attributes;
    :param n: number of users (if known in advance);
    :param f: real frequency of value (if known in advance - not realistic in locally differentially private scenario);
    :return: variance value.
    """

    sig_ue = (1/d) * (d*q + f * (p-q))

    var_ue_z = ((d**2 * sig_ue * (1 - sig_ue)) / (n * (p - q)**2))

    return var_ue_z

def VAR_RSpFD_UE_rnd(p, q, k, d, n=1, f=0):

    """
    Variance value of using RS+FD[UE-rnd] protocol of [1]

    :param p: probability of being 'honest';
    :param q: probability of randomizing value/bit;
    :param k: attribute's domain size;
    :param d: number of attributes;
    :param n: number of users (if known in advance);
    :param f: real frequency of value (if known in advance - not realistic in locally differentially private scenario);
    :return: variance value.
    """

    sig_ue = (1 / d) * (f * (p - q) + q + (d - 1) * ((p / k) + ((k - 1)/(k)) * q) )

    var_ue_rnd = ((d**2 * sig_ue * (1 - sig_ue)) / (n * (p - q)**2))

    return var_ue_rnd
